/*

  fl0p - passive flow fingerprinting
  ----------------------------------

  Q: Know what the difference between your latest project
     and putting wings on an elephant is?

  A: Who knows?  The elephant *might* fly, heh, heh...

  (C) Copyright 2006 by Michal Zalewski <lcamtuf@coredump.cx>

*/

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#ifndef WIN32
#  include <netinet/in.h>
#  include <arpa/inet.h>
#  include <unistd.h>
#  include <netdb.h>
#  include <sys/socket.h>
#  include <sys/un.h>
#  include <pwd.h>
#  include <grp.h>
#else
#  include "getopt.h"
#  include <stdarg.h>
#  pragma comment (lib, "wpcap.lib")
#endif /* ^WIN32 */

#include <stdio.h>
#include <pcap.h>
#include <signal.h>

#ifdef USE_BPF
#include USE_BPF
#else
#include <pcap-bpf.h>
#endif /* ^USE_BPF */

#include <time.h>
#include <ctype.h>

#include "config.h"
#include "types.h"
#include "ip.h"
#include "crc32.h"
#include "fpentry.h"

#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0
#endif /* ! MSG_NOSIGNAL */

static pcap_dumper_t *dumper;

#ifdef WIN32

static inline void debug(_u8* format, ...) {
  _u8 buff[1024];
  va_list args;
  va_start(args, format);
  memset(buff, 0, sizeof(buff));
  _vsnprintf( buff, sizeof(buff) - 1, format, args);
  fprintf(stderr, buff);
  va_end(args);
}

static inline void fatal(_u8* format, ...) {
  _u8 buff[1024];
  va_list args;	
  va_start(args, format);
  memset(buff, 0, sizeof(buff));
  vsnprintf( buff, sizeof(buff) - 1, format, args);
  fprintf(stderr, "[-] ERROR: %s", buff);
  va_end(args);
  exit(1);
}

#else
#  define debug(x...)	fprintf(stderr,x)
#  define fatal(x...)	do { debug("[-] ERROR: " x); exit(1); } while (0)
#endif /* ^WIN32 */

#define pfatal(x)	do { debug("[-] ERROR: "); perror(x); exit(1); } while (0)


static _u8 *config_file,
           *use_iface,
           *use_dump,
#ifndef WIN32
           *set_user,
#endif /* !WIN32 */
           *use_rule = "tcp or udp or icmp";

static _s32 capture_timeout = 1,
            quick_thres = 400;

static _u8 no_known,
           show_unknown,
           no_banner,
           use_promisc,
           add_timestamp,
           header_len,
           no_fuzzy,
           go_daemon,
           use_logfile,
           do_resolve,
           use_concise,
           use_vlan;

static pcap_t *pt;
static struct bpf_program flt;

static _u32 packet_count,
            st_time,
            file_cksum,
            sigcnt;

static struct signature sig[MAXSIGS];
static struct fl0p_stream stream[FLOW_BUCKETS+1][BUCKET_SIZE];


static void die_nicely(_s32 sig) {
  if (sig) debug("+++ Exiting on signal %d +++\n",sig);
  if (pt) pcap_close(pt);
  if (dumper) pcap_dump_close(dumper);

  if (!no_banner && packet_count) {
    float r = packet_count * 60;
    
    r /= (time(0) - st_time);

    debug("[+] Average packet ratio: %0.2f per minute.\n",r);
  }

  exit(sig);
}


static void set_header_len(_u32 type) {

  switch(type) {

    case DLT_SLIP:
    case DLT_RAW:  break;

#ifdef DLT_C_HDLC
    case DLT_C_HDLC:
#endif

    case DLT_NULL: header_len=4; break;

    case DLT_EN10MB: header_len=14; break;

#ifdef DLT_LOOP
    case DLT_LOOP:
#endif

#ifdef DLT_PPP_SERIAL
    case DLT_PPP_SERIAL: /* NetBSD oddity */
#endif

#ifdef DLT_PPP_ETHER
    case DLT_PPP_ETHER:  /* PPPoE on NetBSD */
        header_len=8;
        break;
#endif

    case DLT_PPP:    header_len=4; break;

    case DLT_IEEE802:
      header_len=22;
      break;

#ifdef DLT_IEEE802_11
    case DLT_IEEE802_11: header_len=32; break;
#endif

#ifdef DLT_PFLOG
    case DLT_PFLOG:
      header_len=28;
      break;
#endif

#ifdef DLT_LINUX_SLL
    case DLT_LINUX_SLL:
      header_len=16;
      break;
#endif

    default:
      debug("[!] WARNING: Unknown datalink type %d, assuming no header.\n",type);
      break;

  }

}



static void usage(_u8* name) {
  fprintf(stderr,
          "\nUsage: %s [ -f file ] [ -i device ] [ -s file ] [ -o file ]\n"

#ifndef WIN32
          "       [ -u user ] [ -e ms ] [ -T ms ] [ -FUKrqvpdtl ] [ 'filter rule' ]\n"
#else
          "       [ -e ms ] [ -T ms ] [ -FUKrqvpdLtl ] [ 'filter rule' ]\n"
#endif /* ^WIN32 */

          "  -f file   - read fingerprints from file\n"
          "  -i device - listen on this device\n"
          "  -s file   - read packets from tcpdump snapshot\n"
          "  -o file   - write to this logfile (implies -t)\n"
#ifndef WIN32
          "  -u user   - chroot and setuid to this user\n"
#endif /* !WIN32 */
          "  -e ms     - pcap capture timeout in milliseconds (1)\n"
          "  -q ms     - packet timing threshold in milliseconds (400)\n"
          "  -F        - disable fuzzy matching on all signatures\n"
          "  -U        - display fingerprints for unidentified streams\n"
          "  -K        - do not display known signatures (implies -U)\n"
          "  -r        - resolve host names (not recommended)\n"
          "  -q        - be quiet - no banner\n"
          "  -v        - enable support for 802.1Q VLAN frames\n"
          "  -p        - switch card to promiscuous mode\n"
          "  -d        - daemon mode (fork into background)\n"
#ifdef WIN32
          "  -L        - list all available interfaces\n"
#endif /* ^WIN32 */
          "  -t        - add timestamps to every entry\n"
          "  -l        - output concise 1-line output\n\n"
          "  'Filter rule' is an optional pcap-style BPF expression (man tcpdump).\n\n",name);
  exit(1);
}


/* Not a very strict one - just to catch obvious mistakes... */
static void load_config(_u8* file) {
  _u32 ln=0;
  _u8 buf[MAXLINE];
  _u8* p;
  _u32 elnum;
  _u8  next_delay;
  FILE* c = fopen(file?file:(_u8*)FPRINT_DB,"r");

  if (!c) {
    if (!file) load_config(CONFIG_DIR "/" FPRINT_DB);
      else pfatal(file);
    return;
  }

  while ((p=fgets(buf,sizeof(buf),c))) {
    _u32 l;
    
    elnum = 0;
    next_delay = ITEM_D_NONE;

    file_cksum ^= crc32(buf, strlen(buf));
    ln++;

    /* Remove leading and trailing blanks */
    while (isspace(*p)) p++;
    l=strlen(p);
    while (l && isspace(*(p+l-1))) *(p+(l--)-1)=0;
	
    /* Skip empty lines and comments */
    if (!l) continue;
    if (*p == '#') continue;

    p=strchr(buf,':');
    if (!p) {
      _u32 flag;
      _u8  c;
      if (!(p=strchr(buf,'|')))
        fatal("Missing ':' or '|' in config file (line %d).\n",ln);
      *p = 0;
      p++;
      while (isspace(*p)) p++;
      if (sscanf(p,"%c%i",&c,&flag) != 2 || (c != '+' && c != '-'))
        fatal("Malformed flag set action in config file (line %d).\n",ln);
      if (c == '-') sig[sigcnt].f_set_mode = SIG_FLAG_ZERO; else
        sig[sigcnt].f_set_mode = SIG_FLAG_SET;
      sig[sigcnt].f_set = flag;
    } else {
      *p = 0;
      p++;
      while (isspace(*p)) p++;    
      sig[sigcnt].name = strdup(p);
    }

    p=buf;
    while (isspace(*p)) p++;

    l=strlen(p);
    while (l && isspace(*(p+l-1))) *(p+(l--)-1)=0;

    if (*p == '&' || *p == '!') {
      _u32 flag;
      _u8  c;
      if (sscanf(p,"%c%i",&c,&flag) != 2 || (c != '&' && c != '!'))
        fatal("Malformed flag check action in config file (line %d).\n",ln);
      if (c == '!') sig[sigcnt].f_check_mode = SIG_FLAG_ZERO; else
        sig[sigcnt].f_check_mode = SIG_FLAG_SET;
      sig[sigcnt].f_check = flag;
      p = strchr(p,' ');
      if (!p) fatal("Premature end of signature in config file (line %d).\n",ln);
      while (isspace(*p)) p++;
    }

    
    if (!strncasecmp(p,"TCP ",4)) sig[sigcnt].proto = IPPROTO_TCP;
    else if (!strncasecmp(p,"UDP ",4)) sig[sigcnt].proto = IPPROTO_UDP;
    else if (!strncasecmp(p,"ICMP ",5)) sig[sigcnt].proto = IPPROTO_ICMP;
    else fatal("Invalid protocol in config file (line %d).\n",ln);
    
    p = strchr(p,' ');
    if (!p) fatal("Premature end of signature in config file (line %d).\n",ln);
    while (isspace(*p)) p++;
    if (*p != '*' && !isdigit(*p))
      fatal("Bad port specification in config file (line %d).\n",ln);

    if (*p != '*') {
      _s32 pi = atoi(p);
      if (pi <= 0 || pi > 65535) 
        fatal("Bad port number in config file (line %d).\n",ln);
      sig[sigcnt].port = pi;
    }

    p = strchr(p,' ');
    if (!p) fatal("Premature end of signature in config file (line %d).\n",ln);

    while (isspace(*p) || *p == '=' || *p == '<') {
      if (*p == '=') sig[sigcnt].no_fuzzy = 1;
      if (*p == '<') sig[sigcnt].need_open = 1;
      p++;
    }
    
    while (p && *p) {

      _s32 i, diff = 0;
   
      if (sig[sigcnt].need_close)
        fatal("Trailing garbage after '>' in config file (line %d).\n",ln);
 
      if (*p == '+') { p++; next_delay = ITEM_D_LONG; goto skip_to_next; }
      if (*p == '>') { p++; sig[sigcnt].need_close = 1; goto skip_to_next; }

      if (tolower(*p) == 'c') sig[sigcnt].el[elnum].type = ITEM_CLI;
      else if (tolower(*p) == 's') sig[sigcnt].el[elnum].type = ITEM_SERV;
      else fatal("Unknown packet type '%c' in config file (line %d).\n",*p,ln);
      
      p++;

      switch (*p) {
        case '*':
          sig[sigcnt].el[elnum].valtype = ITEM_VAL_ANY;
          break;

        case '$': 
          sig[sigcnt].el[elnum].valtype = ITEM_VAL_NORM;          
          sig[sigcnt].el[elnum].value = 1075;
          sig[sigcnt].el[elnum].range = 625;
          break;

        case '0' ... '9': 
          sig[sigcnt].el[elnum].valtype = ITEM_VAL_NORM;          
          if (sscanf(p,"%d/%d",&i,&diff) < 1 || i < 0 || i > 65535 
               || diff < 0 || diff > 65535)
            fatal("Bogus packet size / range in config file (line %d).\n",ln);
          sig[sigcnt].el[elnum].value = i;
          sig[sigcnt].el[elnum].range = diff;
          break;
        case '@': 
          sig[sigcnt].el[elnum].valtype = ITEM_VAL_REF;
          p++;
          if (sscanf(p,"%d/%d",&i,&diff) < 1 || i < 1 || i > MAXSIGLEN 
               || diff < 0 || diff > 65535)
            fatal("Bogus packet offset / range in config file (line %d).\n",ln);
          sig[sigcnt].el[elnum].value = i - 1;
          sig[sigcnt].el[elnum].range = diff;
          break;

        default:
          fatal("Bogus packet description in config file (line %d).\n",ln);

      }

      sig[sigcnt].el[elnum].d_prev = next_delay;
      next_delay = ITEM_D_SHORT;

      if (++elnum >= MAXSIGLEN)
        fatal("Maximum signature length (%d) exceeded in config file (line %d).\n",MAXSIGLEN,ln);

      p = strchr(p,' ');

skip_to_next:

      if (p) while (isspace(*p)) p++;
    
    }

    sig[sigcnt].siglen = elnum;

    for (l=0;l<elnum;l++) 
      if (sig[sigcnt].el[elnum].valtype == ITEM_VAL_REF && 
          sig[sigcnt].el[elnum].value >= elnum)
      fatal("Packet reference out of bounds in config file (line %d).\n",ln);
            
    if (++sigcnt >= MAXSIGS)
      fatal("Maximum signature count exceeded.\n");

  }

  fclose(c);

  if (!sigcnt)
    debug("[!] WARNING: no signatures loaded from config file.\n");

}



static void put_date(time_t ut) {
  _u8* x;
  struct tm *tmval;

  switch (add_timestamp) {

    case 0: break;

    case 1: /* localtime */

    case 2: /* UTC */

      x = asctime((add_timestamp == 1) ? localtime(&ut) : 
                                         gmtime(&ut));

      if (x[strlen(x)-1]=='\n') x[strlen(x)-1]=0;

      printf("<%s> ",x);

      break;

    case 3: /* seconds since the epoch */

      printf("<%u> ", (_u32)ut);
      break;

    case 4: /* RFC3339 */
    default:

      tmval = gmtime(&ut);

      printf("<%04u-%02u-%02uT%02u:%02u:%02u> ",
             tmval->tm_year + 1900, tmval->tm_mon + 1, tmval->tm_mday,
             tmval->tm_hour, tmval->tm_min, tmval->tm_sec);

      break;

  }

}


#define MY_MAXDNS 32

static inline _u8* grab_name(_u8* a) {
  struct hostent* r;
  static _u8 rbuf[MY_MAXDNS+6] = "/";
  _u32 j;
  _u8 *s,*d = rbuf + (use_concise == 0);

  if (!do_resolve) return "";

  r = gethostbyaddr(a,4,AF_INET);
  if (!r || !(s = r->h_name) || !(j = strlen(s))) return "";
  if (j > MY_MAXDNS) return "";

  while (j--) {
    if (isalnum(*s) || *s == '-' || *s == '.') *d = *s;
      else *d = '?';
    d++; s++;
  }

  *d=0;

  return rbuf;

}


static _u8* display_bytes(_u64 b) {
  static _u8 retbuf[64];
  if (b < 1024) sprintf(retbuf,"%lluB",b);
  else if (b < 1048576) sprintf(retbuf,"%.01fkB",b/1024.0);
  else if (b < 1073741824) sprintf(retbuf,"%.01fMB",b/1024.0/1024.0);
  else sprintf(retbuf,"%.01fGB",b/1024.0/1024.0/1024.0);
  return retbuf;
}


static void report_match(struct fl0p_stream* f,struct signature* s, _u8 exact) {
  _u8 *x, *a;

  if (no_known) return;

  put_date(f->start_time);
      
  a = (_u8*)&f->caddr;
  x = grab_name(a);

  if (use_concise)
    printf("%s|%u.%u.%u.%u|%s|%u|", (f->proto == IPPROTO_TCP ? "tcp" :
      (f->proto == IPPROTO_UDP ? "udp" : "icmp")), a[0], a[1], a[2], a[3], 
      x, f->cport);
  else 
    printf("(%s) %u.%u.%u.%u%s:%u -> ", (f->proto == IPPROTO_TCP ? "tcp" :
      (f->proto == IPPROTO_UDP ? "udp" : "icmp")), a[0], a[1], a[2], a[3], 
      x, f->cport);

  a = (_u8*)&f->saddr;
  x = grab_name(a);

  if (use_concise)
    printf("%u.%u.%u.%u|%s|%u|0x%x|%llu|%u|%u|%u|%s|%s\n", a[0], a[1], a[2], a[3],
      x, f->sport, f->flag, f->dcount, f->pcount, f->last_time - f->start_time, f->matched,
      exact ? "" : "fuzzy", s->name);
  else printf("%u.%u.%u.%u%s:%u [0x%x]\n"
              "  Observed for: %s, %u packets, spans %d seconds\n"
              "  Matches: %s%s (hit: %u)\n", 
         a[0], a[1], a[2], a[3], x, f->sport, f->flag,
         display_bytes(f->dcount), f->pcount, f->last_time - f->start_time, 
         s->name, exact ? "" : " (fuzzy)", f->matched);

}



static void do_matching(struct fl0p_stream *f,_u8 final) { 

  _u32 i, a, gotone = 0;

  /* Try normal */

  for (i=0;i<sigcnt;i++) {

    if (f->proto != sig[i].proto) continue;
    if (sig[i].port && sig[i].port != f->sport) continue;
    if (sig[i].need_open && f->pcount != sig[i].siglen) continue;
    if (sig[i].need_close && !final) continue;

    if (sig[i].f_check_mode == SIG_FLAG_SET) {
      if ((f->flag & sig[i].f_check) != sig[i].f_check) continue;
    } else if (sig[i].f_check_mode == SIG_FLAG_ZERO)
      if ((f->flag & sig[i].f_check)) continue;

    for (a=0;a<sig[i].siglen;a++) {
      _s32 st = 0, en = 0;
      _u32 tpos = MAXSIGLEN * 2 - sig[i].siglen;

      if (f->tail[tpos + a].type != sig[i].el[a].type) break;
      if (sig[i].el[a].d_prev != ITEM_D_NONE)
        if (f->tail[tpos + a].d_prev != sig[i].el[a].d_prev) break;

      if (sig[i].el[a].valtype == ITEM_VAL_ANY) continue;

      if (sig[i].el[a].valtype == ITEM_VAL_NORM)
        st = en = sig[i].el[a].value;
      else /* ITEM_VAL_REF */
        st = en = f->tail[tpos + sig[i].el[a].value].value;

      st -= sig[i].el[a].range;
      en += sig[i].el[a].range;

      if (f->tail[tpos + a].value < st || f->tail[tpos + a].value > en) break;

    } 

    if (a != sig[i].siglen) continue;

    if (sig[i].name) {
      f->matched++;
      report_match(f,sig+i,1);
      gotone = 1;
    } else {
      if (sig[i].f_set_mode == SIG_FLAG_SET)
        f->flag |= sig[i].f_set;
      else if (sig[i].f_set_mode == SIG_FLAG_ZERO)
        f->flag &= ~sig[i].f_set;
    }

  }

  if (no_fuzzy || gotone) return;

  /* If nothing found, let's do some fuzzy matching... */

  /* No stop-start, normal or reversed C/S... */

  for (i=0;i<sigcnt;i++) {
    _s32 csbal = 0;

    if (sig[i].no_fuzzy) continue;
    if (sig[i].port && sig[i].port != f->sport) continue;

    if (sig[i].f_check_mode == SIG_FLAG_SET) {
      if ((f->flag & sig[i].f_check) != sig[i].f_check) continue;
    } else if (sig[i].f_check_mode == SIG_FLAG_ZERO)
      if ((f->flag & sig[i].f_check)) continue;

    for (a=0;a<sig[i].siglen;a++) {
      _s32 st = 0, en = 0;
      _u32 tpos = MAXSIGLEN * 2 - sig[i].siglen;

      if (f->tail[tpos + a].type == sig[i].el[a].type) break;
      if (f->tail[tpos + a].d_prev == sig[i].el[a].d_prev) csbal++;
        else csbal--;

      if (sig[i].el[a].valtype == ITEM_VAL_ANY) continue;

      if (sig[i].el[a].valtype == ITEM_VAL_NORM)
        st = en = sig[i].el[a].value;
      else /* ITEM_VAL_REF */
        st = en = f->tail[tpos + sig[i].el[a].value].value;

      st -= sig[i].el[a].range;
      en += sig[i].el[a].range;

      if (f->tail[tpos + a].value < st || f->tail[tpos + a].value > en) break;

    }    

    if (abs(csbal) != sig[i].siglen) continue;

    if (sig[i].name) {
      f->matched++;
      report_match(f,sig+i,0);
      gotone = 1;
    } else {
      if (sig[i].f_set_mode == SIG_FLAG_SET)
        f->flag |= sig[i].f_set;
      else if (sig[i].f_set_mode == SIG_FLAG_ZERO)
        f->flag &= ~sig[i].f_set;
    }

  }

  /* TODO: Try ignoring client/server ACKs */

}


#define APPEND_ELEMENT(_f,_type,_value,_delay) \
  do { \
    if (show_unknown && (_f)->pcount < MAXSIGLEN) { \
      memmove((_f)->head,(_f)->head+1, sizeof(struct stream_item) * (MAXSIGLEN-1)); \
      (_f)->head[MAXSIGLEN-1].type = (_type); \
      (_f)->head[MAXSIGLEN-1].value = (_value); \
      (_f)->head[MAXSIGLEN-1].d_prev = (_delay); \
    } \
    memmove((_f)->tail,(_f)->tail+1, sizeof(struct stream_item) * (MAXSIGLEN*2-1)); \
    (_f)->tail[MAXSIGLEN*2-1].type = (_type); \
    (_f)->tail[MAXSIGLEN*2-1].value = (_value); \
    (_f)->tail[MAXSIGLEN*2-1].d_prev = (_delay); \
  } while (0)


static void register_payload(struct fl0p_stream *f, struct timeval* t,_u8 cli, _u16 psize) {
  _u8 dtype = ITEM_D_NONE;

  if (f->pcount) {
    _u32 tdiff = (_s64)t->tv_sec * 1000 + (_s64)t->tv_usec / 1000 -
                 (_s64)f->last_time * 1000 - (_s64)f->last_usec / 1000;
    if (tdiff > quick_thres) dtype = ITEM_D_LONG;
      else dtype = ITEM_D_SHORT;
  }
  
  APPEND_ELEMENT(f,cli ? ITEM_CLI : ITEM_SERV,psize,dtype);

  f->pcount++;
  f->dcount += psize;
  f->last_time = t->tv_sec;
  f->last_usec = t->tv_usec;
  
}


static void finalize_stream(struct fl0p_stream* f, _u8 forced) {
  if (f->proto) {
    do_matching(f,1);
    
    if (!f->matched && show_unknown && f->dcount) {
      _u8 *x, *a, gotany = 0;
      _u32 i;

      put_date(f->start_time);
      
      a = (_u8*)&f->caddr;
      x = grab_name(a);

      if (use_concise)
        printf("%s|%u.%u.%u.%u|%s|%u|", (f->proto == IPPROTO_TCP ? "tcp" :
          (f->proto == IPPROTO_UDP ? "udp" : "icmp")), a[0], a[1], a[2], a[3], 
          x, f->cport);
      else 
        printf("%s %u.%u.%u.%u%s:%u -> ", (f->proto == IPPROTO_TCP ? "tcp" :
          (f->proto == IPPROTO_UDP ? "udp" : "icmp")), a[0], a[1], a[2], a[3], 
          x, f->cport);

      a = (_u8*)&f->saddr;
      x = grab_name(a);

      if (use_concise)
        printf("%u.%u.%u.%u|%s|%u|%x|%llu|%u|%u|%s||UNKNOWN|", a[0], a[1], a[2], a[3],
          x, f->sport, f->flag, f->dcount, f->pcount, f->last_time - f->start_time, forced ? "tmout" : "fin");
      else printf("%u.%u.%u.%u%s:%u [0x%x]\n"
                  "  Totals: %s, %u packets, spans %u seconds, %s\n  UNKNOWN: ",
         a[0], a[1], a[2], a[3], x, f->sport, f->flag,
         display_bytes(f->dcount), f->pcount, f->last_time - f->start_time, forced ? "timeout" : "normal close");
      
      for (i=0;i<MAXSIGLEN;i++) {
        if (f->head[i].type == ITEM_EOF) continue;
        if (f->head[i].d_prev == ITEM_D_LONG) printf(" +");
        if (!gotany) gotany = 1; else putchar(' '); 
        printf("%c%u",f->head[i].type == ITEM_CLI ? 'c' : 's',f->head[i].value); 
      }

      if (!use_concise) {
        printf("\n  Recommended:");
        for (i=0;i<MAXSIGLEN;i++) {
          if (f->head[i].type == ITEM_EOF) continue;
          if (f->head[i].d_prev == ITEM_D_LONG) printf(" +");

          if (f->head[i].value >= 10) {
            _u32 a, ic = 0;
            for (a=0;a<i;a++) {
              if (f->head[a].type == ITEM_EOF) continue;
              ic++;
              if (f->head[a].value == f->head[i].value) {
                printf(" %c@%u",f->head[i].type == ITEM_CLI ? 'c' : 's',ic);
                goto this_one_settled;
              }
            }
          }

          if (f->head[i].value < 10) {
            /* For very small values, exact match might be best */
            printf(" %c%u",f->head[i].type == ITEM_CLI ? 'c' : 's',f->head[i].value); 
          } else if (f->head[i].value < 60) {            
            /* Cover +/- 15% */
            printf(" %c%u/%u",f->head[i].type == ITEM_CLI ? 'c' : 's',f->head[i].value,
              f->head[i].value*15/100); 
          } else if (f->head[i].value < 460) {            
            /* Cover +/- 10 */
            printf(" %c%u/10",f->head[i].type == ITEM_CLI ? 'c' : 's',f->head[i].value); 
          } else if (f->head[i].value < 1600) {
            /* Possible MTU range */
            printf(" %c$",f->head[i].type == ITEM_CLI ? 'c' : 's');
          } else if (f->head[i].value > 65533) {
            printf(" %c%u",f->head[i].type == ITEM_CLI ? 'c' : 's',f->head[i].value);
          } else {
            /* Cover from current -3% to infty */
            _u32 mid = (65535 + (f->head[i].value * 97 / 100)) / 2,
                 ran = (65535 - (f->head[i].value * 97 / 100)) / 2;
            printf(" %c%u/%u",f->head[i].type == ITEM_CLI ? 'c' : 's', mid, ran);
          }

this_one_settled:

        getuid();

        }
      }
      
      printf("\n");

    }
    
    memset(f,0,sizeof(struct fl0p_stream));
  }
}


static void register_packet(_u8 proto, _u32 saddr, _u32 daddr, _u16 sport, 
                            _u16 dport, _s32 psize) {
  _u32 bucket = BUCK(saddr, daddr, sport, dport);
  _u32 i;
  _u32 oldest = 0, oldt = 0xffffffff;
  struct fl0p_stream *f;
  struct timeval t;

  gettimeofday(&t,0);

  for (i=0;i<BUCKET_SIZE;i++) {
    f = &stream[bucket][i];
    if (f->caddr == saddr && f->saddr == daddr &&
        f->cport == sport && f->sport == dport) { /* CLI -> SRV */

      if (psize < 0) { finalize_stream(f,0); return; }	

      /* Timed out - reuse slot */
      if (t.tv_sec - f->last_time > SESS_TIMEOUT) { 
        oldest = i; goto new_stream; 
      }

      if (proto == IPPROTO_TCP && !psize) return; /* Ignore empty ACKs */
      register_payload(f,&t,1,psize);
      do_matching(f,0);
      return;
    } else
    if (f->saddr == saddr && f->caddr == daddr &&
        f->sport == sport && f->cport == dport) { /* SRV -> CLI */

      if (psize < 0) { finalize_stream(f,0); return; }

      /* Timed out - reuse slot */
      if (t.tv_sec - f->last_time > SESS_TIMEOUT) { 
        oldest = i; goto new_stream; 
      }
      
      if (proto == IPPROTO_TCP && !psize) return; /* Ignore empty ACKs */
      register_payload(f,&t,0,psize);
      do_matching(f,0);
      return;
    }

    if (oldt >= f->last_time) { oldest = i; oldt = f->last_time; }

  }

new_stream:

  if (psize < 0) return; /* Whatever - stray RST */

  f = &stream[bucket][oldest];

  finalize_stream(f,1);

  f->proto      = proto;
  f->caddr      = saddr;
  f->saddr      = daddr;
  f->cport      = sport;
  f->sport      = dport;
  f->start_time = t.tv_sec;
  f->last_time  = t.tv_sec;
  
  /* Don't register first no-payload ACK - this is still handshake */
  if (proto != IPPROTO_TCP || psize) register_payload(f,&t,1,psize);

}


static void parse(_u8* none, struct pcap_pkthdr *pph, _u8* packet) {
  struct ip_header   *iph;
  struct tcp_header  *tcph;
  struct udp_header  *udph;
  struct icmp_header *icmph;

  _u8    *end_ptr, *d;

  _s32   ilen, payload;
  _u16   sport, dport;

  packet_count++;

  /* Paranoia! */
  if (pph->len <= PACKET_SNAPLEN) end_ptr = packet + pph->len;
    else end_ptr = packet + PACKET_SNAPLEN;

  iph = (struct ip_header*)(packet+header_len);

  if (use_vlan && iph->ihl == 0x00) 
    iph = (struct ip_header*)((_u8*)iph + 4);

  /* Whoops, IP header ends past end_ptr */
  if ((_u8*)(iph + 1) > end_ptr) return;

  if ( ((iph->ihl & 0x40) != 0x40) || 
       ( iph->proto != IPPROTO_TCP && iph->proto != IPPROTO_UDP &&
         iph->proto != IPPROTO_ICMP) ) {
    debug("[!] WARNING: Non-IP packet received. Bad header_len!\n");
    return;
  }

  /* If the declared length is shorter than the snapshot (etherleak
     or such), truncate this bad boy. */

  d = (_u8*)iph + htons(iph->tot_len);
  if (end_ptr > d) end_ptr = d;

  ilen = iph->ihl & 15;

  /* Borken packet */
  if (ilen < 5) return;

  d = (_u8*)iph + (ilen << 2);

  /* We can't handle fragmented packets well at this point, skip them */
  if (ntohs(iph->off) & IP_MF) return;

  if (iph->proto == IPPROTO_TCP) {
    tcph = (struct tcp_header*) d;
    if (tcph + 1 > (struct tcp_header*) end_ptr) return; /* Whoopsie */
    
    ilen = ((tcph->doff) << 2);
    d += ilen;
    payload = end_ptr - d;

    sport = ntohs(tcph->sport);
    dport = ntohs(tcph->dport);

    if (tcph->flags & (TH_RST|TH_FIN)) payload = -1;
      else if (tcph->flags & TH_SYN) return;
    
  } else if (iph->proto == IPPROTO_UDP) {
    udph = (struct udp_header*) d;
    if (udph + 1 > (struct udp_header*) end_ptr) return; /* Whoopsie */
    payload = end_ptr - (_u8*)(udph+1);
    sport = ntohs(udph->sport);
    dport = ntohs(udph->dport);
  } else {
    icmph = (struct icmp_header*) d;
    if (icmph + 1 > (struct icmp_header*) end_ptr) return; /* Whoopsie */
    payload = end_ptr - (_u8*)(icmph+1);
    sport = dport = 0;
  }

  register_packet(iph->proto,iph->saddr,iph->daddr,sport,dport,payload);
  fflush(0);

}



int main(int argc,char** argv) {
  _u8 buf[MAXLINE*4];
  _s32 r;
  _u8 errbuf[PCAP_ERRBUF_SIZE];

#ifdef WIN32
  _u8 ebuf[PCAP_ERRBUF_SIZE];
  pcap_if_t *alldevs, *d;
  _s32 adapter, i;
  while ((r = getopt(argc, argv, "f:i:s:o:e:t:FUKrqvpdLtlh")) != -1)
#else

  if (getuid() != geteuid())
    fatal("This program is not intended to be setuid.\n");

  while ((r = getopt(argc, argv, "f:i:s:o:u:e:T:FUKrqvpdtlh")) != -1) 
#endif /* ^WIN32 */

    switch (r) {

      case 'f': config_file = optarg; break;

      case 'i': use_iface = optarg; break;

      case 's': use_dump = optarg; break;

      case 'o': if (!freopen(optarg,"a",stdout)) pfatal(optarg);
                use_logfile = 1;
                break;

      case 'e': capture_timeout = atoi(optarg);
                if (capture_timeout <= 0 ||capture_timeout > 10000) fatal("Invalid -e value.\n");
                break;

      case 'T': quick_thres = atoi(optarg);
                if (quick_thres <= 0 || quick_thres > 60000) fatal("Invalid -t value.\n");
                break;

#ifndef WIN32
      case 'u': set_user   = optarg; break;
#endif /* !WIN32 */

      case 'r': do_resolve    = 1; break;
      case 'U': show_unknown  = 1; break;
      case 'K': no_known      = 1; show_unknown = 1; break;
      case 'q': no_banner     = 1; break;
      case 'p': use_promisc   = 1; break;
      case 't': add_timestamp++;   break;
      case 'd': go_daemon     = 1; break;
      case 'v': use_vlan      = 1; break;
      case 'F': no_fuzzy      = 1; break;
      case 'l': use_concise   = 1; break;

#ifdef WIN32

      case 'L':
        if (pcap_findalldevs(&alldevs, ebuf) == -1)
	  fatal("pcap_findalldevs: %s\n", ebuf);

      debug("\nInterface\tDevice\t\tDescription\n"
            "-------------------------------------------\n");

      for(i=1,d=alldevs;d;d=d->next,i++) {
        debug("%d %s",i, d->name);
        if (d->description)
	  debug("\t%s",d->description);
 	debug("\n");
      }
      exit(1);
      break;

#endif  /* WIN32 */

      default: usage(argv[0]);
    }
    
  if (use_logfile && !add_timestamp) add_timestamp = 1;

  if (use_iface && use_dump)
    fatal("-s and -i are mutually exclusive.\n");

  if (!use_logfile && go_daemon)
    fatal("-d requires -o.\n");

  if (!no_banner) {
    debug("fl0p - passive flow fingerprinter, version " VER "\n"
          "(C) Michal Zalewski <lcamtuf@coredump.cx>\n");  
  }

  load_config(config_file);

  if (!show_unknown && !sigcnt)
    fatal("with no signatures in config file, -U is necessary.\n");

  if (argv[optind] && *(argv[optind])) {
    sprintf(buf,"(%s) and (%.3000s)",use_rule,argv[optind]);
    use_rule = buf;
  } 

  if (use_vlan) {
    _u8* x = strdup(use_rule);
    sprintf(buf,"(%.1000s) or (vlan and (%.1000s))",x,x);
    free(x);
    use_rule = buf;
  }

  signal(SIGINT,&die_nicely);
  signal(SIGTERM,&die_nicely);

#ifndef WIN32
  signal(SIGHUP,&die_nicely);
  signal(SIGQUIT,&die_nicely);
#endif /* !WIN32 */

  if (use_dump) {
    if (!(pt=pcap_open_offline(use_dump, errbuf))) 
      fatal("pcap_open_offline failed: %s\n",errbuf);
  } else {

#ifdef WIN32
    if (pcap_findalldevs(&alldevs, ebuf) == -1)
      fatal("pcap_findalldevs: %s\n", ebuf);
	
    if (!use_iface) {
      d = alldevs;
    } else {
      adapter = atoi(use_iface);
      for(i=1, d=alldevs; adapter && i < adapter && d; i++, d=d->next);
      if (!d) fatal("Unable to find adapter %d\n", adapter);
    }

    use_iface = d->name;

#else
    if (!use_iface) use_iface=pcap_lookupdev(errbuf);
#endif /* ^WIN32 */

    if (!use_iface) use_iface = "lo";

    /* We do not rely on pcap timeouts - they suck really bad. Of
       course, the documentation sucks, and if you use the timeout
       of zero, things will break. */
    
    if (!(pt=pcap_open_live(use_iface,PACKET_SNAPLEN,use_promisc,capture_timeout,errbuf))) 
      fatal("pcap_open_live failed: %s\n",errbuf);
  }

  set_header_len(pcap_datalink(pt));

  if (pcap_compile(pt, &flt, use_rule, 1, 0))
    if (strchr(use_rule,'(')) {
      pcap_perror(pt,"pcap_compile");
      debug("See man tcpdump or p0f README for help on bpf filter expressions.\n");
      exit(1);
    }

  if (!no_banner) {
    debug("fl0p: listening on '%s', %d sigs (cksum %08X), rule: '%s'.\n",
          use_dump?use_dump:use_iface,sigcnt,file_cksum,
          argv[optind]?argv[optind]:"all");

  }
  
  pcap_setfilter(pt, &flt);

#ifndef WIN32

  if (set_user) {
    struct passwd* pw;

    if (geteuid()) fatal("only root can use -u.\n");

    tzset();

    pw = getpwnam(set_user);
    if (!pw) fatal("user %s not found.\n",set_user);
    
    if (chdir(pw->pw_dir)) pfatal(pw->pw_dir);
    if (chroot(pw->pw_dir)) pfatal("chroot");
    chdir("/");

    if (initgroups(pw->pw_name,pw->pw_gid)) pfatal("initgroups");
    if (setgid(pw->pw_gid)) pfatal("setgid");
    if (setuid(pw->pw_uid)) pfatal("setuid");

    if (getegid() != pw->pw_gid || geteuid() != pw->pw_uid)
      fatal("failed to setuid/setgid to the desired UID/GID.\n");

  }

#endif /* !WIN32 */

  if (go_daemon) {

#ifndef WIN32
    _s32 f;
    FILE* pid_fd;
    fflush(0);
    f = fork();
    if (f<0) pfatal("fork() failed");
    if (f) exit(0);
    dup2(1,2);
    close(0);
    chdir("/");
    setsid();
    signal(SIGHUP,SIG_IGN);
    
    if ((pid_fd = fopen(PID_PATH, "w"))) {
      fprintf(pid_fd, "%d", getpid());
      fclose(pid_fd);
    }
    
    printf("--- fl0p " VER " willing to serve again (");
    put_date(time(0));
    printf(") ---\n");
    fflush(0);
#else
    fatal("daemon mode is not supported in the WIN32 version.\n");
#endif /* ^WIN32 */

  }

  st_time = time(0);

  pcap_loop(pt,-1,(pcap_handler)&parse,0);

  pcap_close(pt);
  if (dumper) pcap_dump_close(dumper);

  if (use_dump) debug("[+] End of input file.\n");
    else fatal("Network is down.\n");

  return 0;

}

